"""The primary yalchemy interface

Consult the ``from_dict`` method of the yalchemy objects for information
on how they are constructed.
"""
# Geoalchemy must be imported for geography points to be
# reflected properly
import geoalchemy2

from yalchemy.pretty_yaml import PrettyYalchemyDumper
from yalchemy.table_structures import Column
from yalchemy.table_structures import Index
from yalchemy.table_structures import ForeignKey
from yalchemy.table_structures import UniqueConstraint
from yalchemy.table_structures import CheckConstraint
from yalchemy.table_structures import Table
from yalchemy.table_structures import TableSet
from yalchemy.version import __version__  # flake8: noqa


__all__ = [
    'Column',
    'Index',
    'ForeignKey',
    'UniqueConstraint',
    'CheckConstraint',
    'Table',
    'TableSet',
    'PrettyYalchemyDumper',
]



""" Unit tests for column utilities """

import pytest

from yalchemy import col_utils


@pytest.mark.parametrize('datatype,expected', [
    ('int8', 'bigint'),
    ('integer []', 'integer[]'),
    ('date []', 'date[]'),
    ('float4', 'real'),
    ('timestamp with time zone', 'timestamp_with_time_zone', ),
    ('timestamp   with   time zone', 'timestamp_with_time_zone'),
    ('bool', 'boolean'),
    ])
def test_clean_datatype(expected, datatype):
    """ Test that datatypes that we recieve are cleaned and
    are not alias names in postgres """

    assert col_utils.clean_datatype(datatype) == expected


def test_is_schema_qualified():
    assert col_utils.is_schema_qualified('my_schema.my_sequence') is True
    assert col_utils.is_schema_qualified('my_sequence') is False


@pytest.mark.parametrize('input_string, expected_output', [
    # various quotes styles
    ("'test'", 'test'),
    ('"test"', 'test'),
    ('\'"test"\'', 'test'),
    # normal string
    ('test', 'test'),
])
def test_unquote_sql_string(input_string, expected_output):
    assert col_utils.unquote_sql_string(input_string) == expected_output



"""Common table abstractions and utilities."""
# pylint: disable=arguments-differ,too-many-lines
import abc
import collections
import copy
import enum
import functools
import hashlib
import re

import sqlalchemy as sa
import sqlalchemy.sql.elements as sa_elements
import sqlalchemy.sql.functions as sa_func
import sqlalchemy.dialects.postgresql as sa_pg
import yaml

from yalchemy import col_utils, pretty_yaml, sqla_utils

# TODO: validation - duplicate names, invalid foreign keys, etc.


class Error(Exception):
    """The base exception class for this module."""


class MergeError(Error):
    """Raised when trying to merge two incompatible table abstractions."""


class NoDataType(Error):
    """ Raised when there is no data type given for a column """


class InvalidColumnDefault(Error):
    """ Raised when an invalid column default is specified """


class _ComparableMixin:
    """Mixin to provide common comparator functionality."""
    __slots__ = ()

    def _as_tuple(self):
        return tuple(getattr(self, attr) for attr in self.__slots__)

    def _comparable(self, other):
        return all(hasattr(other, attr) for attr in self.__slots__)

    def __eq__(self, other):
        if not self._comparable(other):  # pragma: no cover
            return NotImplemented
        else:
            return self._as_tuple() == other._as_tuple()

    def __ror__(self, other):
        """Implement reverse or to handle `None | object => object`."""
        if other is None:
            return self
        else:  # pragma: no cover
            return NotImplemented


try:
    DefaultYamlLoader = yaml.CSafeLoader
except AttributeError:  # pragma: no cover
    DefaultYamlLoader = yaml.SafeLoader


class Yalchemy(metaclass=abc.ABCMeta):
    """ The base yalchemy object that defines the interface for serialization """

    @abc.abstractmethod
    def _to_dict(self):
        """ Converts a yalchemy object to a dictionary """

    def to_dict(self):
        """ Converts a yalchemy object to a dictionary.

        .. note::

            Wrapper around _to_dict that will order the _to_dict
            returned dict by the __slots__ order in an OrderedDict.
            Please remember to override ``_to_dict`` and not ``to_dict``
            when subclassing yalchemy.
        """
        unordered_dict = self._to_dict()
        ordered_dict = collections.OrderedDict()
        for key in self.__slots__:
            if key in unordered_dict:
                ordered_dict[key] = unordered_dict.pop(key)
        assert not unordered_dict
        return ordered_dict

    @abc.abstractclassmethod
    def from_dict(cls, dict_obj):
        """ Constructs a yalchemy object from a dictionary """

    @abc.abstractmethod
    def to_sqla(self, **kwargs):
        """ Converts a yalchemy object to a sqlalchemy object

        Args:
            kwargs: Keyword arguments specific to the yalchemy object
                being created. Some objects need special parameters for their
                construction (e.g. sqlalchemy MetaData)
        """

    @classmethod
    @abc.abstractmethod
    def from_sqla(cls, sqla_obj):
        """ Constructs a yalchemy object from a sqlalchemy object

        Args:
            sqla_obj: The sqlalchemy object
        """

    def to_yaml(self):
        """ Converts a yalchemy object to a yaml string """
        return yaml.dump(self.to_dict(), Dumper=pretty_yaml.PrettyYalchemyDumper)

    @classmethod
    def from_yaml(cls, yaml_str, loader_factory=None):
        """ Loads a yalchemy object from a yaml string

        Args:
            yaml (str): The yaml string
            loader_factory (function(stream)): (optional) constructs a custom PyYaml loader.
                If not provided and libyaml is installed, it will default to CSafeLoader.
                If not provided and libyaml is not installed, it will default to SafeLoader.
        """
        loader_factory = loader_factory or DefaultYamlLoader
        loader = loader_factory(yaml_str)

        try:
            dict_obj = loader.get_single_data()
        finally:
            loader.dispose()

        if not isinstance(dict_obj, dict):  # pragma: no cover
            raise ValueError('from_yaml only supports loading yaml dictionaries')

        return cls.from_dict(dict_obj)

    def __deepcopy__(self, memo):
        cls = self.__class__
        result = cls.__new__(cls)
        memo[id(self)] = result
        for key in self.__slots__:
            setattr(result, key, copy.deepcopy(getattr(self, key), memo))
        return result


class ColumnDefaultType(enum.Enum):
    sequence = 'sequence'
    expression = 'expression'


class ColumnDefault(Yalchemy, _ComparableMixin):
    __slots__ = ('type', 'value')
    REGEX_SEQUENCE_EXPRESSION = re.compile(
        r'^nextval\((?P<sequence>[\w\.\'\"\\]+)(?:::regclass)?\)$'
    )

    def __init__(self, type_, value):
        assert isinstance(type_, ColumnDefaultType)
        assert isinstance(value, str)

        if type_ == ColumnDefaultType.sequence and \
                not col_utils.is_schema_qualified(value):
            raise InvalidColumnDefault('Schema must be specified for sequence types')

        self.type = type_
        self.value = col_utils.unquote_sql_string(value)

    @classmethod
    def from_dict(cls, dict_obj):
        """ Creates a yalchemy ColumnDefault from a dictionary.

        This is for SQLAlchemy database (server) defaults
        http://docs.sqlalchemy.org/en/latest/core/metadata.html#sqlalchemy.schema.Column.params.server_default
        https://www.postgresql.org/docs/current/static/ddl-default.html

        The dictionary has the following specification::

        {
            # The type of the default (expression or sequence)
            'type': string(required),

            # DDL clause for default
            #
            # examples:
            #   'NOW()', 'UUID_GENERATE_V4()', ... for functions
            #   'schema.sequence_name' for sequences (must include schema)
            #   '1', 'foo' for quoted_string (includes numeric values)
            'value': str(required)
        }

        """
        return cls(ColumnDefaultType(dict_obj['type']), dict_obj['value'])

    @classmethod
    def from_sqla(cls, sqla_obj):
        # extract object from default clause
        if isinstance(sqla_obj, sa.DefaultClause):
            sqla_obj = sqla_obj.arg

        # extract sequence object from a next_value() clause
        if isinstance(sqla_obj, sa_func.next_value):
            sqla_obj = sqla_obj.sequence

        # convert SQLAlchemy object to appropriate ColumnDefaultType
        if isinstance(sqla_obj, sa.Sequence):
            default_type = ColumnDefaultType.sequence
            if not sqla_obj.schema:
                raise InvalidColumnDefault(
                    'Sequence {} must be qualified with a schema'.format(str(sqla_obj)),
                )
            value = '{}.{}'.format(sqla_obj.schema, str(sqla_obj.name))
        elif isinstance(sqla_obj, sa_elements.TextClause):
            # determine if the text clause is actually a sequence and extract it
            sequence_match = cls.REGEX_SEQUENCE_EXPRESSION.match(sqla_obj.text)
            if sequence_match:
                default_type = ColumnDefaultType.sequence
                value = col_utils.unquote_sql_string(sequence_match.group('sequence'))
            else:
                default_type = ColumnDefaultType.expression
                value = sqla_obj.text
        elif isinstance(sqla_obj, str):
            default_type = ColumnDefaultType.expression
            value = col_utils.unquote_sql_string(sqla_obj)
        else:
            raise TypeError(
                'Server default object must be a sequence, string or SQLAlchemy TextClause')

        return cls(default_type, value)

    def _to_dict(self):
        return {
            'type': self.type.value,
            'value': self.value,
        }

    def to_sqla(self):
        if self.type == ColumnDefaultType.sequence:
            sequence_schema, sequence_name = self.value.split('.')
            return sa.Sequence(sequence_name, schema=sequence_schema)
        else:
            return sa.text(self.value)

    def __or__(self, other):
        if not (self.type == other.type and
                self.value == other.value):
            raise MergeError(
                'Cannot merge ColumnDefaults: {0!r}, {1!r}'.format(self, other))
        return type(self)(self.type, self.value)

    def __copy__(self):
        return ColumnDefault(self.type, self.value)

    def __repr__(self):
        return ('ColumnDefault(type={s.type!r}, '
                'value={s.value!r})'.format(s=self))


class Column(Yalchemy, _ComparableMixin):
    """ A yalchemy Column that can be converted to a sqlalchemy Column """
    __slots__ = ('name', 'datatype', 'format', 'required', 'default', 'doc')

    def __init__(self, name, datatype, format_=None, required=False, default=None, doc=None):
        self.name = str(name)
        self.datatype = col_utils.clean_datatype(datatype)
        self.format = format_
        self.required = required
        self.default = default

        # doc isn't read from sqlalchemy objects or written to them
        # since it's not supported yet (will be in version 1.2)
        self.doc = doc

    @classmethod
    def from_dict(cls, dict_obj):
        """ Creates a yalchemy Column from a dictionary.

        The dictionary has the following specification::

            {
                # The name of the column
                'name': str(required),

                # The datatype. Can be any of the postgres datatypes listed
                # at https://www.postgresql.org/docs/current/static/datatype.html#DATATYPE-TABLE
                'datatype': str(required),

                # The format of the datatype. This is an array of arguments that is passed
                # into the datatype. For example, a format of [255] for a varchar type would
                # create a varchar(255) column datatype
                'format': list(default=[]),

                # If the datatype is required. If False, it will be null
                'required': bool(default=False),

                # The database server-side default for the column. Column defaults follow the
                # specification provided in the `ColumnDefault.from_dict` documentation
                'default': dict(optional)
            }

        """
        dict_obj = copy.copy(dict_obj)
        if 'datatype' not in dict_obj:
            raise NoDataType('No datatype given for col: %s' % dict_obj.get('name'))
        dict_obj['format_'] = dict_obj.pop('format', None)
        if 'default' in dict_obj:
            dict_obj['default'] = ColumnDefault.from_dict(dict_obj['default'])
        else:
            dict_obj['default'] = None
        return cls(**dict_obj)

    @classmethod
    def from_sqla(cls, sqla_obj):
        """ Create a yalchemy Column using a sqlalchemy Column

        Args:
            sqla_obj (Column): The sqlalchemy column
        """
        col_type, col_format = cls._get_col_type(sqla_obj.type)
        # Column formats default to [] with sqlalchemy, however, yalchemy defaults them to None
        col_format = None if col_format == [] else col_format
        if sqla_obj.server_default:
            col_default = ColumnDefault.from_sqla(sqla_obj.server_default)
        else:
            col_default = None

        return cls(str(sqla_obj.name), col_type, col_format,
                   required=not sqla_obj.nullable, default=col_default)

    def _to_dict(self):
        """ Converts the yalchemy Column to a dictionary """

        metadata = {
            'name': self.name,
            'datatype': self.datatype,
            'required': self.required,
        }

        if self.doc:
            metadata['doc'] = self.doc
        if self.format:
            metadata['format'] = self.format
        if self.default:
            metadata['default'] = self.default.to_dict()

        return metadata

    def to_sqla(self):  # pylint: disable=arguments-differ
        """ Converts the yalchemy Column to a sqlalchemy Column """
        sa_type = sqla_utils.sa_type_from_str(self.datatype, self.format)
        column_args = [self.name, sa_type]

        if self.default:
            default_sqla_obj = self.default.to_sqla()
            # sequences
            if isinstance(default_sqla_obj, sa.Sequence):
                column_args.append(default_sqla_obj)
                server_default = default_sqla_obj.next_value()
            else:
                server_default = default_sqla_obj
        else:
            server_default = None

        return sa.Column(
            *column_args, key=self.name, nullable=not self.required,
            server_default=server_default)

    @staticmethod
    def _get_col_type(sqlalchemy_type):
        """ Use the sqlalchemy type and compile it into
        it's ddl. We then lowercase it and make the spaces underscores
        because that looks prettier in the yaml. """
        clean_value = sqlalchemy_type.compile(sa_pg.dialect()).lower()
        # get's the raw type and any args it takes
        # for example it splits varchar(255) into varchar and 255
        matches = re.match(r'^([\w ]*)(\((.*)\))?(\[\])?$', clean_value)
        col_type = matches.group(1) + (matches.group(4) or '')
        if matches.group(3):
            col_format = [int(i) if i != 'point' else i for i in matches.group(3).split(',')]
        else:
            col_format = []
        return col_type.replace(' ', '_'), col_format

    def __or__(self, other):
        # TODO(d-felix): Support merging of unequal datatypes, e.g. date and datetime.
        if other is None:
            return self
        elif not self._comparable(other):  # pragma: no cover
            return NotImplemented

        if not (self.name == other.name and
                self.datatype == other.datatype and
                self.format == other.format and
                self.default == other.default and
                self.doc == other.doc):
            raise MergeError(
                'Cannot merge Columns: {0!r}, {1!r}'.format(self, other))

        return type(self)(
            name=self.name,
            datatype=self.datatype,
            format_=self.format,
            required=self.required and other.required,
            default=self.default,
            doc=self.doc)

    def __copy__(self):
        return Column(self.name, self.datatype,
                      format_=self.format, required=self.required,
                      default=self.default, doc=self.doc)

    def __repr__(self):
        return ('Column(name={s.name!r}, datatype={s.datatype!r}, '
                'format={s.format!r}, required={s.required!r}, doc={s.doc!r}, '
                'default={s.default!r})'.format(s=self))


@functools.total_ordering  # make ForeignKey sortable via __eq__ and __lt__
class ForeignKey(Yalchemy, _ComparableMixin):
    """ A yalchemy ForeignKey that can be converted to a sqlalchemy ForeignKeyIndex """
    __slots__ = ('column', 'remote_table', 'remote_column')

    def __init__(self, column, remote_table, remote_column):
        if isinstance(column, (list, tuple)):
            self.column = [str(col) for col in column]
        else:
            self.column = str(column)
        self.remote_table = remote_table
        self.remote_column = remote_column

    @classmethod
    def from_dict(cls, dict_obj):
        """ Creates a yalchemy ForeignKey from a dictionary.

        The dictionary has the following specification::

            {
                # The name of the column
                'column': str(required),

                # The name of the remote table
                'remote_table': str(required),

                # The name of the remote column
                'remote_column': str(required),
            }

        .. note:: The ``column`` and ``remote_column`` arguments can
            be list of strings in the case of creating foreign keys on
            combinations of columns.
        """
        return cls(**dict_obj)

    @classmethod
    def from_sqla(cls, sqla_obj):
        """ Create a yalchemy ForeignKey from a sqlalchemy ForeignKeyConstraint

        Arguments:
            sqla_obj: The sqlalchemy ForeignKeyConstraint object.
        """

        assert len(sqla_obj.column_keys) == 1
        assert len(sqla_obj.elements) == 1

        target_table, target_col = sqla_obj.elements[0].target_fullname.rsplit('.', 1)
        return cls(sqla_obj.column_keys[0], target_table, target_col)

    def _to_dict(self):
        """ Converts the yalchemy Foreignkey to a dictionary """
        return {
            'column': self.column,
            'remote_table': self.remote_table,
            'remote_column': self.remote_column,
        }

    def to_sqla(self):  # pylint: disable=arguments-differ
        """ Converts the yalchemy ForeignKey to a sqlalchemy ForeignKeyConstraint """

        to_column = self.remote_table + '.' + self.remote_column
        return sa.ForeignKeyConstraint(
            [self.column], [to_column], deferrable=True, initially='IMMEDIATE')

    def __lt__(self, other):  # pragma: no cover
        if not self._comparable(other):
            return NotImplemented
        return self._as_tuple() < other._as_tuple()  # pylint: disable=protected-access

    def __hash__(self):
        return hash(self._as_tuple())

    def __copy__(self):  # pragma: no cover
        return ForeignKey(self.column, self.remote_table, self.remote_column)

    def __repr__(self):  # pragma: no cover
        return ('ForeignKey(column={s.column!r}, remote_table={s.remote_table!r}, '
                'remote_column={s.remote_column!r})'.format(s=self))


@functools.total_ordering  # make sortable via __eq__ and __lt__
class _ColumnCollection(Yalchemy, _ComparableMixin):
    """ Base class for Yalchemy index column collections """
    __slots__ = ('name', 'columns',)
    HASHED_NAME_PREFIX = None

    def __init__(self, columns, name=None):
        self.columns = columns
        self.name = name

    @classmethod
    def from_dict(cls, dict_obj):
        """ Creates a yalchemy column collection from a dictionary.

        The dictionary has the following specification::

            {
                # The list of column names. If multiple are
                # provided, a multidimensional index will be created
                'columns': list[str](required),

                # A fixed name for this object
                # (if not specified, will automatically generate a hashed name)
                'name': str(optional)
            }
        """
        return cls(dict_obj['columns'], name=dict_obj.get('name'))

    def _to_dict(self):
        """ Converts the column collection to a dictionary """

        result = {'columns': self.columns}

        # include name only if it was explicitly set
        if self.name:
            result['name'] = self.name

        return result

    def to_sqla(self, table_name):  # pylint: disable=arguments-differ
        """ Converts the yalchemy column collection to a sqlalchemy object.

        Args:
            table_name (str): The name of the table to which the sqlalchemy object is applied
        """
        return self._make_sqla_object(
            self.name if self.name else self._create_hashed_name(table_name)
        )

    def _create_hashed_name(self, table_name):
        """ Create the hashed name of this collection object for the table given the list
        of columns and the table name """

        params_str = table_name + ':' + ','.join(self.columns)
        hash_value = hashlib.md5(params_str.encode('utf-8')).hexdigest()[0:8]

        return '{prefix}__{hash_value}__{table_name}'.format(
            prefix=self.HASHED_NAME_PREFIX,
            hash_value=hash_value,
            table_name=table_name
        )[0:63]

    @abc.abstractmethod
    def _make_sqla_object(self, name):
        """ Constructs the sqlalchemy object associated with this yalchemy object """

    def _as_tuple(self):
        """Compare only on the columns, not the name."""
        return tuple(self.columns)

    def __lt__(self, other):
        if not self._comparable(other):  # pragma: no cover
            return NotImplemented
        return self._as_tuple() < other._as_tuple()  # pylint: disable=protected-access

    def __hash__(self):
        return hash(self._as_tuple())

    def __copy__(self):  # pragma: no cover
        cls = self.__class__
        return cls(self.columns, name=self.name)

    def __repr__(self):
        name_detail = ', name={!r}'.format(self.name) if self.name else ''
        column_detail = ', '.join(repr(c) for c in self.columns)
        return '{}(columns=[{}]{})'.format(self.__class__.__name__, column_detail, name_detail)


class Index(_ColumnCollection):
    """ A yalchemy Index that can be converted to a sqlalchemy Index """
    HASHED_NAME_PREFIX = 'ix'

    def _make_sqla_object(self, name):
        return sa.Index(name, *self.columns)

    @classmethod
    def from_sqla(cls, sqla_obj):
        """ Create a yalchemy Index from a sqlalchemy Index """
        return cls(
            columns=[str(c.name if isinstance(c, sa.Column) else c)
                     for c in sqla_obj.expressions])


class UniqueConstraint(_ColumnCollection):
    """ A yalchemy UniqueConstraint that can be converted to a sqlalchemy UniqueConstraint """
    HASHED_NAME_PREFIX = 'uq'

    @classmethod
    def from_sqla(cls, sqla_obj):
        """ Create a yalchemy UniqueConsraint from a sqlalchemy UniqueConstraint """
        return cls(
            columns=[str(c.name) for c in sqla_obj.columns])

    def _make_sqla_object(self, name):
        return sa.UniqueConstraint(*self.columns, name=name)


@functools.total_ordering  # make Check constraint sortable via __eq__ and __lt__
class CheckConstraint(Yalchemy, _ComparableMixin):
    """ A yalchemy CheckConstraint that can be converted to a sqlalchemy CheckConstraint

    .. note::

        While other yalchemy objects perform equality comparisons on all attributes
        of their classes, CheckConstraint does not do comparison on the 'check' attribute.
        This is because Postgres may format the check value when storing it.
    """
    __slots__ = ('name', 'check')

    def __init__(self, name, check):
        self.name = name
        self.check = check

    @classmethod
    def from_dict(cls, dict_obj):
        """ Creates a yalchemy CheckConstraint from a dictionary.

        The dictionary has the following specification::

            {
                # The check is the string of the check we want to
                # have on the constraint. It should be verbatium
                # what the check would be in SQL
                'name': str(required),
                'check': str(required),
            }
        """
        return cls(name=dict_obj['name'], check=dict_obj['check'])

    @classmethod
    def from_sqla(cls, sqla_obj):
        """ Create a yalchemy CheckConstraint from a sqlalchemy CheckConstraint """

        # handles a weird case where sqlalchemy names check constraints
        # without a name with this name instead of having it be None
        if sqla_obj.name is None or str(sqla_obj.name) == '_unnamed_':
            name = None
        else:
            name = str(sqla_obj.name)
        return cls(name, str(sqla_obj.sqltext))

    def _to_dict(self):
        """ Converts the yalchemy CheckConstraint to a dictionary """
        return {'name': self.name, 'check': self.check}

    def to_sqla(self):  # pylint: disable=arguments-differ
        """ Converts the yalchemy CheckConstraint to a sqlalchemy CheckConstraint. """

        return sa.CheckConstraint(self.check, name=self.name)

    def _as_tuple(self):
        """Convert to tuple for ease of comparison.

        .. note::

            The "check" parameter has been removed from comparison since it can be
            reflected differently from postgres once it is stored. In other words, if
            check constraints have the same name, they are considered identical.
        """
        return (self.name,)

    def __lt__(self, other):  # pragma: no cover
        if not self._comparable(other):
            return NotImplemented
        return self._as_tuple() < other._as_tuple()  # pylint: disable=protected-access

    def __hash__(self):
        return hash(self._as_tuple())

    def __copy__(self):  # pragma: no cover
        return CheckConstraint(self.name, self.check)

    def __repr__(self):  # pragma: no cover
        return 'CheckConstraint(name={s.name!r}, check={s.check!r})'.format(s=self)


class Table(Yalchemy, _ComparableMixin):  # pylint: disable=too-many-instance-attributes
    """ A yalchemy Table that can be converted to a sqlalchemy Table """
    __slots__ = ('name', 'schema', 'columns', 'primary_keys', 'foreign_keys', 'indexes',
                 'unique_constraints', 'check_constraints', 'doc')

    def __init__(self, name, schema, columns=None, foreign_keys=None,
                 primary_keys=None, indexes=None, check_constraints=None, unique_constraints=None,
                 doc=None):
        self.name = str(name)
        self.schema = str(schema)
        self.columns = columns or ()
        self.foreign_keys = set(foreign_keys or ())
        self.indexes = set(indexes or ())
        self.primary_keys = [str(key) for key in primary_keys or []]
        self.check_constraints = set(check_constraints or ())
        self.unique_constraints = set(unique_constraints or ())
        self.doc = doc

    @property
    def full_name(self):
        """ Get the full name of the table along with the schema name """
        return self.schema + '.' + self.name

    @classmethod
    def from_dict(cls, dict_obj):
        """ Creates a yalchemy Table from a dictionary.

        The dictionary has the following specification::

            {
                # The table name
                'name': str(required),

                # The schema
                'schema': str(required),

                # The list of table column dictioniaries. Column dictionaries
                # follow the specification provided in the `Column.from_dict` documentation
                'columns': list[dict](optional),

                # The list of column foreign key dictionaries. Foreign keys follow the
                # specification provided in the `ForeignKey.from_dict` documentation
                'foreign_keys': list[dict](optional),

                # The list of index dictionaries. Indexes follow the
                # specification provided in the `Index.from_dict` documentation
                'indexes': list[dict](optional),

                # The list of column names (as strings) that are primary keys. The unique
                # constraint is made with the columns in the given order.
                'primary_keys': list[str](optional),

                # The list of unique constraints.  Unique constraints follow the
                # specification provided in the `UniqueConstraint.from_dict` documentation

                # The list of check constraints.  Check constraints follow the
                # specification provided in the `CheckConstraint.from_dict` documentation
                'check_constraints': list[dict](optional)

                # Documentation for the table. Not read from sqlalchemy objs
                # or written to them until SA version 1.2
                'doc': str(optional)
            }

        A full example of a table looks like this::

            {
                'name': 'my_table',
                'schema': 'schema',
                'doc': 'Some document string'
                'columns': [
                    {'name': 'col1', 'datatype': 'varchar', 'format': [123], 'required': False},
                    {'name': 'col2', 'datatype': 'integer', 'required': True,
                     'doc': 'col2 is a good column'},
                ],
                'foreign_keys': [
                    {'column': 'col2', 'remote_table': 'other_table', 'remote_column': 'other_col'}
                ],
                'indexes': [{'columns': ['col1', 'col2']}],
                'unique_constraints': [{'name': 'my_unique_constraint', 'columns': ['col2']}],
                'check_constraints': [{'name': 'my_check', 'check': 'char_length(col1) = col2'}]
                'primary_keys': ['col2']
            }
        """
        dict_obj = copy.copy(dict_obj)
        dict_obj['columns'] = [
            Column.from_dict(c) for c in dict_obj.get('columns', [])
        ]
        dict_obj['foreign_keys'] = [
            ForeignKey.from_dict(f) for f in dict_obj.get('foreign_keys', [])
        ]
        dict_obj['indexes'] = [
            Index.from_dict(i) for i in dict_obj.get('indexes', [])
        ]
        dict_obj['unique_constraints'] = [
            UniqueConstraint.from_dict(i) for i in dict_obj.get('unique_constraints', [])
        ]
        dict_obj['check_constraints'] = [
            CheckConstraint.from_dict(i) for i in dict_obj.get('check_constraints', [])
        ]

        return cls(**dict_obj)

    @classmethod
    def from_sqla(cls, sqla_obj):
        """ Create a yalchemy Table from a sqlalchemy Table.

        Arguments:
            sqla_obj: The sqlalchemy Table object
        """

        columns = [Column.from_sqla(col) for col in sqla_obj.c]
        foreign_keys = [ForeignKey.from_sqla(fkey.constraint) for col in sqla_obj.c
                        for fkey in col.foreign_keys]
        primary_keys = [col.name for col in sqla_obj.primary_key]
        indexes = [Index.from_sqla(i) for i in sqla_obj.indexes]

        check_constraints = [CheckConstraint.from_sqla(c) for c in sqla_obj.constraints
                             if isinstance(c, sa.CheckConstraint)]
        unique_constraints = [UniqueConstraint.from_sqla(c) for c in sqla_obj.constraints
                              if isinstance(c, sa.UniqueConstraint)]
        return cls(
            sqla_obj.name, schema=sqla_obj.schema, columns=columns, foreign_keys=foreign_keys,
            primary_keys=primary_keys, indexes=indexes, check_constraints=check_constraints,
            unique_constraints=unique_constraints)

    def _to_dict(self):
        """ Converts the yalchemy Table to a dictionary """
        columns = [c.to_dict() for c in self.columns]

        table_metadata = {
            'name': self.name,
            'schema': self.schema,
            'columns': columns,
        }
        if self.doc:
            table_metadata['doc'] = self.doc
        if self.primary_keys:
            table_metadata['primary_keys'] = self.primary_keys
        if self.foreign_keys:
            table_metadata['foreign_keys'] = [fk.to_dict() for fk in sorted(self.foreign_keys)]
        if self.indexes:
            table_metadata['indexes'] = [ix.to_dict() for ix in sorted(self.indexes)]
        if self.unique_constraints:
            table_metadata['unique_constraints'] = [
                c.to_dict() for c in sorted(self.unique_constraints)]
        if self.check_constraints:
            table_metadata['check_constraints'] = [
                c.to_dict() for c in sorted(self.check_constraints)]
        return table_metadata

    def to_sqla(self, metadata=None, include_indexes=True):  # pylint: disable=arguments-differ
        """ Converts the yalchemy Table to a sqlalchemy Table.

        Args:
            metadata (MetaData): The sqlalchemy metdata object. If ``None``, creates a new
                sqlalchemy MetaData instance.
            include_indexes (bool, default=True): Include indexes when creating the
                sqlalchemy Table object
        """
        metadata = metadata or sa.MetaData()
        indexes = []
        unique_constraints = []
        if include_indexes:
            indexes = [i.to_sqla(table_name=self.name) for i in self.indexes]
            unique_constraints = [c.to_sqla(table_name=self.name) for c in self.unique_constraints]
        check_constraints = [c.to_sqla() for c in self.check_constraints]
        columns = [c.to_sqla() for c in self.columns]
        primary_key = [sa.PrimaryKeyConstraint(*self.primary_keys)]
        table_elements = columns + primary_key + indexes + unique_constraints + check_constraints

        table = sa.Table(
            self.name,
            metadata,
            schema=self.schema,
            *table_elements)
        for fkey_obj in self.foreign_keys:
            fkey = fkey_obj.to_sqla()
            table.append_constraint(fkey)
        return table

    def __or__(self, other):
        if other is None:
            return self
        if not self._comparable(other):
            return NotImplemented

        if not (self.name == other.name and self.primary_keys == other.primary_keys):
            raise MergeError(
                'Cannot merge Tables: {0!r}, {1!r}'.format(self, other))

        name_to_column = collections.OrderedDict((c.name, c) for c in self.columns)
        for column in other.columns:
            column_name = column.name
            name_to_column[column_name] = name_to_column.get(column_name) | column

        return type(self)(
            name=self.name,
            schema=self.schema,
            columns=list(name_to_column.values()),
            foreign_keys=self.foreign_keys & other.foreign_keys,
            indexes=self.indexes | other.indexes,
            primary_keys=self.primary_keys,
            unique_constraints=self.unique_constraints & other.unique_constraints,
            check_constraints=self.check_constraints & other.check_constraints,
            doc=self.doc)

    def __copy__(self):  # pragma: no cover
        return Table(
            self.name,
            self.schema,
            columns=self.columns,
            foreign_keys=self.foreign_keys,
            primary_keys=self.primary_keys,
            indexes=self.indexes,
            unique_constraints=self.unique_constraints,
            check_constraints=self.check_constraints,
            doc=self.doc)

    def __repr__(self):
        return 'Table(name={s.name!r}, primary_keys={s.primary_keys!r})'.format(s=self)


class TableSet(Yalchemy, _ComparableMixin):
    """ A yalchemy TableSet that can be converted to a list of sqlalchemy Tables """
    __slots__ = ('tables',)

    def __init__(self, tables=None):
        self.tables = tables or []

    @classmethod
    def from_dict(cls, dict_obj):
        """ Creates a yalchemy TableSet from a dictionary.

        The dictionary has the following specification::

            {
                # The list of table dictionaries. The specification for a table
                # dictionary can be found in the `Table.from_dict` documentation.
                'tables': list[dict](optional)
            }
        """
        return cls([Table.from_dict(t) for t in dict_obj['tables']])

    @classmethod
    def from_sqla(cls, sqla_objs):  # pylint: disable=arguments-differ
        """ Create a yalchemy TableSet from a list of sqlalchemy Tables.

        Arguments:
            sqla_objs (list[Table]): A list of sqlalchemy tables
        """

        return cls(tables=[Table.from_sqla(t) for t in sqla_objs])

    def _to_dict(self):
        """ Converts the yalchemy TableSet to a dictionary """
        return {
            'tables': [t.to_dict() for t in self.tables],
        }

    # pylint: disable=arguments-differ
    def to_sqla(self, metadata=None, include_indexes=True,
                add_unique_constraints=True):
        """ Converts the yalchemy TableSet to a list of sqlalchemy Tables.

        Arguments:
            include_indexes (bool, default=True): Include the indexes in the
                returned sqlalchemy tables.
            add_unique_constraints (bool, default=True):  Add unique constraints
                to the targets of foreign keys. Requires the target tables of foreign
                keys to be in this table set.
        """

        if not metadata:
            metadata = sa.MetaData()

        tables = {}
        table_to_unique_columns = collections.defaultdict(set)
        for table in self.tables:
            sqla_table = table.to_sqla(
                metadata=metadata, include_indexes=include_indexes)
            tables[table.full_name] = sqla_table
            for foreign_key in table.foreign_keys:
                table_to_unique_columns[foreign_key.remote_table].add(
                    foreign_key.remote_column)

        if add_unique_constraints:
            for remote_table, constrains in table_to_unique_columns.items():
                for constraint in constrains:
                    tables[remote_table].append_constraint(
                        sa.UniqueConstraint(constraint))

        return list(tables.values())

    def __or__(self, other):
        if other is None:
            return self
        elif not self._comparable(other):  # pragma: no cover
            return NotImplemented

        name_to_table = collections.OrderedDict((t.name, t) for t in self.tables)
        for table in other.tables:
            table_name = table.name
            name_to_table[table_name] = name_to_table.get(table_name) | table

        return type(self)(tables=list(name_to_table.values()))

    def __copy__(self):  # pragma: no cover
        return TableSet(self.tables)



"""Tests for clover.data_ingest.parsing.parsers.table_structures"""
# pylint: disable=too-many-lines
import copy

import pytest
import sqlalchemy as sa
import sqlalchemy.dialects.postgresql as sa_pg
import sqlalchemy.sql.elements as sa_elements
import sqlalchemy.sql.functions as sa_func

from yalchemy import table_structures


# COLUMN DEFAULTS

def test_column_default_from_dict():
    """ Test that we build this ColumnDefault object correctly from a dict """
    col_default = table_structures.ColumnDefault.from_dict(
        {'type': 'expression', 'value': 'foobar'}
    )
    assert col_default.type == table_structures.ColumnDefaultType.expression
    assert col_default.value == 'foobar'

    col_default = table_structures.ColumnDefault.from_dict(
        {'type': 'expression', 'value': 'NOW()'},
    )
    assert col_default.type == table_structures.ColumnDefaultType.expression
    assert col_default.value == 'NOW()'

    col_default = table_structures.ColumnDefault.from_dict(
        {'type': 'sequence', 'value': 'schema.id_seq'},
    )
    assert col_default.type == table_structures.ColumnDefaultType.sequence
    assert col_default.value == 'schema.id_seq'

    with pytest.raises(table_structures.InvalidColumnDefault) as exc:
        table_structures.ColumnDefault.from_dict(
            {'type': 'sequence', 'value': 'unqualified_seq'},
        )
    assert 'Schema must be specified for sequence types' in str(exc.value)


@pytest.mark.parametrize('sqla_server_default, expected_default_type, expected_value', [
    # unquoted string
    ('foobar',
     table_structures.ColumnDefaultType.expression, 'foobar'),
    # quoted strings
    ('"foobar"',
     table_structures.ColumnDefaultType.expression, 'foobar'),
    ("'foobar'",
     table_structures.ColumnDefaultType.expression, 'foobar'),
    # standard expression
    (sa.text('NOW()'),
     table_structures.ColumnDefaultType.expression, 'NOW()'),
    # plain Sequence
    (sa.Sequence('id_seq', schema='schema'),
     table_structures.ColumnDefaultType.sequence, 'schema.id_seq'),
    # sequences resulting from SQLAlchemy internals or table reflection
    (sa.text('nextval(\'"schema.primary_key_seq"\'::regclass)'),
     table_structures.ColumnDefaultType.sequence, 'schema.primary_key_seq'),
    (sa.text("nextval('schema.primary_key_seq'::regclass)"),
     table_structures.ColumnDefaultType.sequence, 'schema.primary_key_seq'),
    (sa.text("nextval('schema.primary_key_seq')"),
     table_structures.ColumnDefaultType.sequence, 'schema.primary_key_seq'),
    (sa_func.next_value(sa.Sequence('primary_key_seq', schema='schema')),
     table_structures.ColumnDefaultType.sequence, 'schema.primary_key_seq'),
])
def test_column_default_from_sqla(sqla_server_default, expected_default_type,
                                  expected_value):
    col_default = table_structures.ColumnDefault.from_sqla(sqla_server_default)
    assert col_default.type == expected_default_type
    assert col_default.value == expected_value


def test_column_default_from_invalid_sqla():
    with pytest.raises(TypeError) as exc:
        table_structures.ColumnDefault.from_sqla(1)
    assert 'must be a sequence, string or SQLAlchemy TextClause' in str(exc.value)

    with pytest.raises(table_structures.InvalidColumnDefault) as exc:
        table_structures.ColumnDefault.from_sqla(sa.text("nextval('primary_key_seq')"))
    assert 'Schema must be specified for sequence types' in str(exc.value)


@pytest.mark.parametrize('column ,expected_metadata', [
    (table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression, '1'),
     {'type': 'expression', 'value': '1'}),
    (table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression, "'foobar'"),
     {'type': 'expression', 'value': 'foobar'}),
    (table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression, 'NOW()'),
     {'type': 'expression', 'value': 'NOW()'}),
    (table_structures.ColumnDefault(table_structures.ColumnDefaultType.sequence,
                                    'schema.my_col_seq'),
     {'type': 'sequence', 'value': 'schema.my_col_seq'}),
])
def test_column_default_to_dict(column, expected_metadata):
    assert column.to_dict() == expected_metadata


def test_column_default_to_sqla():
    col_default = table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression,
                                                 'foobar')
    sa_obj = col_default.to_sqla()
    assert isinstance(sa_obj, sa_elements.TextClause)
    assert str(sa_obj) == 'foobar'

    col_default = table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression,
                                                 'NOW()')
    sa_obj = col_default.to_sqla()
    assert isinstance(sa_obj, sa_elements.TextClause)
    assert str(sa_obj) == 'NOW()'

    col_default = table_structures.ColumnDefault(
        table_structures.ColumnDefaultType.sequence,
        'schema.id_seq'
    )
    sa_obj = col_default.to_sqla()
    assert isinstance(sa_obj, sa.Sequence)
    assert sa_obj.name == 'id_seq'
    assert sa_obj.schema == 'schema'


@pytest.mark.parametrize('left, right, expected', [
    # same parameters
    (table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression, 'NOW()'),
     table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression, 'NOW()'),
     table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression, 'NOW()')),
    # different default types
    pytest.mark.xfail(
        (table_structures.ColumnDefault(table_structures.ColumnDefaultType.sequence,
                                        'myschema.my_seq'),
         table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression,
                                        'myschema.my_seq'),
         None),
        raises=table_structures.MergeError, strict=True),
    # different default expressions
    pytest.mark.xfail(
        (table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression, 'foo'),
         table_structures.ColumnDefault(table_structures.ColumnDefaultType.expression, 'bar'),
         None),
        raises=table_structures.MergeError, strict=True),
])
def test_column_default_or(left, right, expected):
    assert (left | right) == expected


def test_column_default_copy():
    column_default = table_structures.ColumnDefault(
        table_structures.ColumnDefaultType.expression, 'foobar'
    )
    copy1 = copy.copy(column_default)
    assert copy1 == column_default
    assert copy1 is not column_default


# COLUMNS


def test_column_from_dict():
    """ Test that we build this Column object correctly from a dict """

    col = table_structures.Column.from_dict(
        {'name': 'col1', 'datatype': 'text', 'format': 'abc', 'required': True})

    assert col.name == 'col1'
    assert col.datatype == 'text'
    assert col.format == 'abc'
    assert col.required is True
    assert col.default is None

    col = table_structures.Column.from_dict(
        {'name': 'col1', 'datatype': 'text', 'required': False})

    assert col.name == 'col1'
    assert col.datatype == 'text'
    assert col.format is None
    assert col.required is False
    assert col.default is None

    col = table_structures.Column.from_dict(
        {'name': 'col1', 'datatype': 'text', 'required': True,
         'default': {'value': 'barfoo', 'type': 'expression'}},
    )

    assert col.name == 'col1'
    assert col.datatype == 'text'
    assert col.format is None
    assert col.required is True
    assert col.default == table_structures.ColumnDefault(
        table_structures.ColumnDefaultType.expression, 'barfoo')

    col = table_structures.Column.from_dict(
        {'name': 'col1', 'datatype': 'timestamptz', 'required': True,
         'default': {'value': 'NOW()', 'type': 'expression'}},
    )

    assert col.name == 'col1'
    assert col.datatype == 'timestamp_with_time_zone'
    assert col.format is None
    assert col.required is True
    assert isinstance(col.default, table_structures.ColumnDefault)
    assert col.default.type == table_structures.ColumnDefaultType.expression
    assert col.default.value == 'NOW()'

    col = table_structures.Column.from_dict(
        {'name': 'col1', 'datatype': 'bigint', 'required': True,
         'default': {'value': 'schema.primary_key_seq', 'type': 'sequence'}},
    )

    assert col.name == 'col1'
    assert col.datatype == 'bigint'
    assert col.format is None
    assert col.required is True
    assert isinstance(col.default, table_structures.ColumnDefault)
    assert col.default.type == table_structures.ColumnDefaultType.sequence
    assert col.default.value == 'schema.primary_key_seq'


def test_column_from_sqla():
    """ Test that we take a sqlalchemy.Column and make the yalchemy column """
    col = table_structures.Column.from_sqla(
        sa.Column('col1', sa.VARCHAR(255), primary_key=True, nullable=False))

    assert col.name == 'col1'
    assert col.datatype == 'varchar'
    assert col.format == [255]
    assert col.required is True
    assert col.default is None

    col = table_structures.Column.from_sqla(
        sa.Column('col1', sa.Integer, nullable=True))

    assert col.name == 'col1'
    assert col.datatype == 'integer'
    assert col.format is None
    assert col.required is False
    assert col.default is None

    col = table_structures.Column.from_sqla(
        sa.Column('col1', sa.Integer, nullable=False,
                  # SQLAlchemy requires all server-side default primitives to be strings
                  server_default='1'))

    assert col.name == 'col1'
    assert col.datatype == 'integer'
    assert col.format is None
    assert col.required is True
    assert col.default == table_structures.ColumnDefault(
        table_structures.ColumnDefaultType.expression, '1')

    col = table_structures.Column.from_sqla(
        sa.Column('col1', sa_pg.UUID(), nullable=False,
                  server_default=sa.text('uuid_generate_v4()')))

    assert col.name == 'col1'
    assert col.datatype == 'uuid'
    assert col.format is None
    assert col.required is True
    assert col.default == table_structures.ColumnDefault(
        table_structures.ColumnDefaultType.expression, 'uuid_generate_v4()')

    # pylint: disable=no-value-for-parameter
    test_sequence = sa.Sequence('primary_key_seq', schema='schema')
    col = table_structures.Column.from_sqla(
        sa.Column('col1', sa.BigInteger(),
                  test_sequence,
                  server_default=test_sequence.next_value(),
                  nullable=False))
    # pylint: enable=no-value-for-parameter

    assert col.name == 'col1'
    assert col.datatype == 'bigint'
    assert col.format is None
    assert col.required is True
    assert col.default == table_structures.ColumnDefault(
        table_structures.ColumnDefaultType.sequence, 'schema.primary_key_seq')

    # should fail without schema
    test_sequence = sa.Sequence('primary_key_seq')
    with pytest.raises(table_structures.InvalidColumnDefault) as exc:
        # pylint: disable=no-value-for-parameter
        table_structures.Column.from_sqla(
            sa.Column('col1', sa.BigInteger(),
                      test_sequence,
                      server_default=test_sequence.next_value(),
                      nullable=False))
        # pylint: enable=no-value-for-parameter
    assert 'must be qualified with a schema' in str(exc.value)


@pytest.mark.parametrize('column, expected_metadata', [
    (table_structures.Column(name='foo', datatype='string', doc='foo doc'),
     {'name': 'foo', 'datatype': 'string', 'required': False, 'doc': 'foo doc'}),
    (table_structures.Column(name='foo', datatype='string', required=True),
     {'name': 'foo', 'datatype': 'string', 'required': True}),
    (table_structures.Column(name='bar', datatype='boolean', format_='t|f'),
     {'name': 'bar', 'datatype': 'boolean', 'format': 't|f', 'required': False}),
    (table_structures.Column(name='zing', datatype='string', required=True,
                             default=table_structures.ColumnDefault(
                                 table_structures.ColumnDefaultType.expression,
                                 'zang')),
     {'name': 'zing', 'datatype': 'string', 'required': True,
      'default': {'value': 'zang', 'type': 'expression'}}),
    (table_structures.Column(name='zing', datatype='timetz', required=True,
                             default=table_structures.ColumnDefault(
                                 table_structures.ColumnDefaultType.expression,
                                 'current_time')),
     {'name': 'zing', 'datatype': 'time_with_time_zone', 'required': True,
      'default': {'value': 'current_time', 'type': 'expression'}}),

    (table_structures.Column(name='zang', datatype='bigint', required=True,
                             default=table_structures.ColumnDefault(
                                 table_structures.ColumnDefaultType.sequence,
                                 'schema.primary_key_seq')),
     {'name': 'zang', 'datatype': 'bigint', 'required': True,
      'default': {'value': 'schema.primary_key_seq', 'type': 'sequence'}}),
])
def test_column_to_dict(column, expected_metadata):
    assert column.to_dict() == expected_metadata


@pytest.mark.parametrize(
    'yalchemy_col,sa_col,default_sa_obj_cls,default_sa_expression', [
        (table_structures.Column(name='my_col', datatype='integer', required=False),
         sa.Column('my_col', sa.Integer, nullable=True),
         None,
         None),
        (table_structures.Column(name='my_col', datatype='varchar', format_=[123], required=True),
         sa.Column('my_col', sa.VARCHAR(123), nullable=False),
         None,
         None),
        (table_structures.Column(name='my_col', datatype='uuid', format_=[True], required=True,
                                 default=table_structures.ColumnDefault(
                                     table_structures.ColumnDefaultType.expression,
                                     'uuid_generate_v4()')),
         sa.Column('my_col', sa_pg.UUID(as_uuid=True), nullable=False,
                   server_default=sa.text('uuid_generate_v4()')),
         sa_elements.TextClause,
         'uuid_generate_v4()'),
        (table_structures.Column(name='my_col', datatype='integer', required=True,
                                 default=table_structures.ColumnDefault(
                                     table_structures.ColumnDefaultType.expression,
                                     '1')),
         sa.Column('my_col', sa.Integer, nullable=False, server_default='1'),
         sa_elements.TextClause,
         '1'),
        (table_structures.Column(name='my_col', datatype='integer', required=True,
                                 default=table_structures.ColumnDefault(
                                     table_structures.ColumnDefaultType.sequence,
                                     'schema.my_col_seq')),
         # pylint: disable=no-value-for-parameter
         sa.Column('my_col', sa.Integer, sa.Sequence('my_col_seq', schema='schema'),
                   nullable=False,
                   server_default=sa.Sequence('my_col_seq', schema='schema').next_value()),
         # pylint: enable=no-value-for-parameter
         sa_func.next_value,
         'schema.my_col_seq'),
    ]
)
def test_column_to_sqla(yalchemy_col, sa_col, default_sa_obj_cls, default_sa_expression):
    """ Test that we turn a yalchemy column into a sqlalchemy column """

    generated_col = yalchemy_col.to_sqla()
    assert generated_col.name == sa_col.name
    assert generated_col.type.compile(sa_pg.dialect()) == \
        sa_col.type.compile(sa_pg.dialect())
    assert generated_col.nullable == sa_col.nullable
    if default_sa_obj_cls is not None:
        assert isinstance(generated_col.server_default, sa.DefaultClause)

        wrapped_server_default = generated_col.server_default.arg
        assert isinstance(wrapped_server_default, default_sa_obj_cls)

        if default_sa_obj_cls == sa_elements.TextClause:
            value = wrapped_server_default.text
        else:
            assert default_sa_obj_cls == sa_func.next_value
            value = '{}.{}'.format(
                wrapped_server_default.sequence.schema,
                wrapped_server_default.sequence.name,
            )
        assert value == default_sa_expression


@pytest.mark.parametrize('left, right, expected', [
    (table_structures.Column('foo', 'string'),
     table_structures.Column('foo', 'string'),
     table_structures.Column('foo', 'string')),
    (table_structures.Column('foo', 'string'),
     None,
     table_structures.Column('foo', 'string')),
    (None,
     table_structures.Column('foo', 'string'),
     table_structures.Column('foo', 'string')),
    pytest.mark.xfail(
        (table_structures.Column('foo', 'string'),
         table_structures.Column('bar', 'string'),
         table_structures.Column('foo', 'string')),
        raises=table_structures.MergeError, strict=True),
    pytest.mark.xfail(
        (table_structures.Column('foo', 'string'),
         table_structures.Column('foo', 'integer'),
         table_structures.Column('foo', 'string')),
        raises=table_structures.MergeError, strict=True),
    (table_structures.Column('foo', 'boolean', format_='t|f'),
     table_structures.Column('foo', 'boolean', format_='t|f'),
     table_structures.Column('foo', 'boolean', format_='t|f')),
    pytest.mark.xfail(
        (table_structures.Column('foo', 'boolean', format_='t|f'),
         table_structures.Column('foo', 'boolean', format_='1|0'),
         table_structures.Column('foo', 'boolean', format_='t|f')),
        raises=table_structures.MergeError, strict=True),
    (table_structures.Column(name='foo', datatype='string',
                             default={'value': 'foo', 'type': 'expression'}),
     table_structures.Column(name='foo', datatype='string',
                             default={'value': 'foo', 'type': 'expression'}),
     table_structures.Column(name='foo', datatype='string',
                             default={'value': 'foo', 'type': 'expression'})),
    # different default expressions
    pytest.mark.xfail(
        (table_structures.Column(name='foo', datatype='string',
                                 default={'value': 'foo', 'type': 'expression'}),
         table_structures.Column(name='foo', datatype='string',
                                 default={'value': 'bar', 'type': 'expression'}),
         table_structures.Column(name='foo', datatype='string',
                                 default={'value': 'NOTPOSSIBLE', 'type': 'expression'})),
        raises=table_structures.MergeError, strict=True),
    # # left - no default, right - has default
    pytest.mark.xfail(
        (table_structures.Column(name='foo', datatype='string'),
         table_structures.Column(name='foo', datatype='string',
                                 default={'value': 'foo', 'type': 'expression'}),
         table_structures.Column(name='foo', datatype='string',
                                 default={'value': 'NOTPOSSIBLE', 'type': 'expression'})),
        raises=table_structures.MergeError, strict=True),
], ids=str)
def test_column_or(left, right, expected):
    assert (left | right) == expected


def test_column_copy():
    column = table_structures.Column('foo', 'boolean', [], False)
    copy1 = copy.copy(column)
    assert copy1 == column
    assert copy1 is not column
    copy2 = copy.deepcopy(column)
    assert copy2 == column
    assert copy2 is not column
    assert copy2.format is not column.format


# FOREIGN KEYS


def test_foreign_key_from_dict():
    """ Test that we get a yalchemy foreign key from a dict correctly """

    fkey = table_structures.ForeignKey.from_dict(
        {'column': 'user_id', 'remote_table': 'user', 'remote_column': 'users'})

    assert fkey.column == 'user_id'
    assert fkey.remote_table == 'user'
    assert fkey.remote_column == 'users'


def test_foreign_key_from_sqla():
    """ Test that we take a sqlalchemy.ForeignKeyConstraints and
    make the yalchemy foreign key """

    fkey = table_structures.ForeignKey.from_sqla(
        sa.ForeignKeyConstraint(['test_col'], ['other_table.other_col']))

    assert fkey.column == 'test_col'
    assert fkey.remote_table == 'other_table'
    assert fkey.remote_column == 'other_col'


@pytest.mark.parametrize('fkey, expected_metadata', [
    (table_structures.ForeignKey('other_foo', 'other', 'foo'),
     {'column': 'other_foo',
      'remote_table': 'other', 'remote_column': 'foo'}),
    (table_structures.ForeignKey(['other_foo', 'other_bar'], 'other', ['foo', 'bar']),
     {'column': ['other_foo', 'other_bar'],
      'remote_table': 'other', 'remote_column': ['foo', 'bar']}),
])
def test_foreign_key_to_dict(fkey, expected_metadata):
    assert fkey.to_dict() == expected_metadata


def test_foreign_key_to_sqla():
    """ Test that we make the ForeignKeyConstraint correctly in sqlalchemy """

    fkey_obj = table_structures.ForeignKey('test_col', 'other_table', 'other_col')
    fkey = fkey_obj.to_sqla()

    assert isinstance(fkey, sa.ForeignKeyConstraint)
    assert len(fkey.elements) == 1
    assert fkey.column_keys == ['test_col']
    assert fkey.elements[0].target_fullname == 'other_table.other_col'


def test_foreign_key_hasing():
    fkey1 = table_structures.ForeignKey('foo', 'bar', 'baz')
    fkey2 = table_structures.ForeignKey('foo', 'bar', 'baz')
    assert {fkey1: 1}[fkey2] == 1
    assert {fkey1} == {fkey2}
    assert {fkey1, fkey2} == {fkey1}


# INDEXES

@pytest.mark.parametrize('idx, expected_metadata', [
    (table_structures.Index(['col_1']),
     {'columns': ['col_1']}),
    (table_structures.Index(['col_1', 'col_2']),
     {'columns': ['col_1', 'col_2']}),
])
def test_index_to_dict(idx, expected_metadata):
    assert idx.to_dict() == expected_metadata


def test_index_from_dict():
    """ Test that we get a yalchemy Index from a dict correctly """

    index = table_structures.Index.from_dict(
        {'columns': ['col1', 'col2']})

    assert index.columns == ['col1', 'col2']


def test_index_from_sqla():
    """ Test that we take a sqlalchemy.Index and make the yalchemy Index """

    index = table_structures.Index.from_sqla(
        sa.Index('some_index', 'a_col', 'another_col'))

    assert index.columns == ['a_col', 'another_col']


def test_index_to_sqla_unnamed():
    """ Test that we make the sa.Index correctly from a yalchemy Index """

    index_obj = table_structures.Index(columns=['col1', 'col2'])
    index = index_obj.to_sqla(table_name='123')

    assert isinstance(index, sa.Index)
    # this is the correct hash for this table + column names
    assert index.name == 'ix__18122589__123'
    assert set(index.expressions) == {'col1', 'col2'}


def test_index_to_sqla_named():
    """ Test that we make the sa.Index correctly from a yalchemy Index """

    index_obj = table_structures.Index(columns=['col1', 'col2'], name='my_index')
    index = index_obj.to_sqla(table_name='123')

    assert isinstance(index, sa.Index)
    assert index.name == 'my_index'
    assert set(index.expressions) == {'col1', 'col2'}


def test_index_hashing():
    idx1 = table_structures.Index(['col_1', 'col_2'])
    idx2 = table_structures.Index(['col_1', 'col_2'])
    assert {idx1: 1}[idx2] == 1
    assert {idx1} == {idx2}
    assert {idx1, idx2} == {idx1}


def test_index_str_repr():
    idx1 = table_structures.Index(['col_1', 'col_2'])
    assert str(idx1) == "Index(columns=['col_1', 'col_2'])"

    idx2 = table_structures.Index(['col_1', 'col_2'], name='my_fixed_name')
    assert str(idx2) == "Index(columns=['col_1', 'col_2'], name='my_fixed_name')"


# UNIQUE CONSTRAINTS

@pytest.mark.parametrize('constraint, expected_metadata', [
    (table_structures.UniqueConstraint(['col_1']),
     {'columns': ['col_1']}),
    (table_structures.UniqueConstraint(['col_2', 'col_1']),
     {'columns': ['col_2', 'col_1']}),
])
def test_unique_to_dict(constraint, expected_metadata):
    assert constraint.to_dict() == expected_metadata


def test_unique_from_dict():
    """ Test that we get a yalchemy UniqueConstraint from a dict correctly """

    constraint = table_structures.UniqueConstraint.from_dict(
        {'columns': ['col1', 'col2']})

    assert constraint.columns == ['col1', 'col2']


def test_unique_from_sqla():
    """ Test that we take a sqlalchemy.UniqueConstraint and make the yalchemy UniqueConstraint """

    # unique constraint needs to be bound to a table
    sa_table = sa.Table(
        'test_table',
        sa.MetaData(),
        sa.Column('a_col', sa.Integer, primary_key=True),
        sa.Column('another_col', sa.Text),
        sa.UniqueConstraint('a_col', 'another_col', name='some_constraint'),
        schema='test_schema')
    unique_constraint = next(
        c for c in sa_table.constraints if isinstance(c, sa.UniqueConstraint)
    )  # pragma: no cover

    constraint = table_structures.UniqueConstraint.from_sqla(unique_constraint)

    assert constraint.columns == ['a_col', 'another_col']


def test_unique_to_sqla_unnamed():
    """ Test that we make the sa.UniqueConstraint correctly from a yalchemy UniqueConstraint """
    constraint_obj = table_structures.UniqueConstraint(columns=['col1', 'col2'])
    constraint = constraint_obj.to_sqla(table_name='123')

    assert isinstance(constraint, sa.UniqueConstraint)
    # this is the correct hash for this table + column names
    assert constraint.name == 'uq__18122589__123'

    # must be bound to a table to verify the resulting columns
    sa_table = sa.Table(  # noqa: F841  # pylint: disable=unused-variable
        'test_table',
        sa.MetaData(),
        sa.Column('col1', sa.Integer),
        sa.Column('col2', sa.Integer),
        constraint,
        schema='test_schema')

    assert {c.name for c in constraint.columns} == {'col1', 'col2'}


def test_unique_to_sqla_named():
    """ Test that we make the sa.UniqueConstraint correctly from a yalchemy UniqueConstraint """
    constraint_obj = table_structures.UniqueConstraint(
        columns=['col1', 'col2'],
        name='my_constraint'
    )
    constraint = constraint_obj.to_sqla(table_name='123')

    assert isinstance(constraint, sa.UniqueConstraint)
    assert constraint.name == 'my_constraint'

    # must be bound to a table to verify the resulting columns
    sa_table = sa.Table(  # noqa: F841  # pylint: disable=unused-variable
        'test_table',
        sa.MetaData(),
        sa.Column('col1', sa.Integer),
        sa.Column('col2', sa.Integer),
        constraint,
        schema='test_schema')

    assert {c.name for c in constraint.columns} == {'col1', 'col2'}


def test_unique_hashing():
    unique1 = table_structures.UniqueConstraint(['col_1', 'col_2'])
    unique2 = table_structures.UniqueConstraint(['col_1', 'col_2'])
    assert {unique1: 1}[unique2] == 1
    assert {unique1} == {unique2}
    assert {unique1, unique2} == {unique1}


def test_unique_str_repr():
    idx1 = table_structures.UniqueConstraint(['col_1', 'col_2'])
    assert str(idx1) == "UniqueConstraint(columns=['col_1', 'col_2'])"

    idx2 = table_structures.UniqueConstraint(['col_1', 'col_2'], name='my_fixed_name')
    assert str(idx2) == "UniqueConstraint(columns=['col_1', 'col_2'], name='my_fixed_name')"


# CheckConstraint


@pytest.mark.parametrize('constraint, expected_metadata', [
    (table_structures.CheckConstraint('check1', 'col1 > col2'),
     {'name': 'check1', 'check': 'col1 > col2'}),
    (table_structures.CheckConstraint('check2', '((col1 == col2))'),
     {'name': 'check2', 'check': '((col1 == col2))'}),
])
def test_constraint_to_dict(constraint, expected_metadata):
    """ Test that we convert a yalchemy CheckConstraint
    into the proper dict format """
    assert constraint.to_dict() == expected_metadata


def test_constraint_from_dict():
    """ Test that we get a yalchemy CheckConstraint from a dict correctly """

    constraint = table_structures.CheckConstraint.from_dict(
        {'name': 'check1', 'check': 'col1 > col2'})

    assert constraint.name == 'check1'
    assert constraint.check == 'col1 > col2'


def test_constraint_from_sqla(transacted_postgresql_db):
    """ Test that we take a sqlalchemy.Column and make the yalchemy CheckConstraint """

    constraint = table_structures.CheckConstraint.from_sqla(
        sa.CheckConstraint('col1 > col2', name='a_check'))

    assert constraint.name == 'a_check'
    assert constraint.check == 'col1 > col2'

    # test from one without a name
    constraint = table_structures.CheckConstraint.from_sqla(
        sa.CheckConstraint('col1 > col2'))

    assert constraint.name is None
    assert constraint.check == 'col1 > col2'

    # test when sql-alchemy reflection calls the name '_unnamed_'
    constraint = table_structures.CheckConstraint.from_sqla(
        sa.CheckConstraint(
            'col1 > col2',
            name=sa.sql.elements._defer_none_name(value='_unnamed_'),
        )
    )
    assert constraint.name is None
    assert constraint.check == 'col1 > col2'


def test_constraint_to_sqla():
    """ Test that we make the sa.CheckConstraint correctly from a yalchemy CheckConstraint """

    constraint_obj = table_structures.CheckConstraint(name='check1', check='col1 < col2')
    constraint = constraint_obj.to_sqla()

    assert isinstance(constraint, sa.CheckConstraint)
    assert constraint.name == 'check1'
    assert str(constraint.sqltext) == 'col1 < col2'


def test_constraint_hasing():
    con1 = table_structures.CheckConstraint('check1', 'col1 > col2')
    con2 = table_structures.CheckConstraint('check1', 'col1 > col2')
    assert {con1: 1}[con2] == 1
    assert {con1} == {con2}
    assert {con1, con2} == {con1}


# TABLES


def test_table_from_dict():
    """ Test that we build a yalchemy Table from a dict """

    table_dict = {
        'name': 'my_table',
        'schema': 'schema',
        'columns': [
            {'name': 'col1', 'datatype': 'varchar', 'format': [123], 'required': False},
            {'name': 'col2', 'datatype': 'integer', 'required': True},
            {'name': 'col3', 'datatype': 'integer', 'required': True,
             'default': {'value': '-1', 'type': 'expression'}},
        ],
        'foreign_keys': [
            {'column': 'col2', 'remote_table': 'other_table', 'remote_column': 'other_col'}
        ],
        'indexes': [{'columns': ['col1', 'col2']}],
        'unique_constraints': [{'columns': ['col3']}],
        'primary_keys': ['col2', 'col1']
    }
    table = table_structures.Table.from_dict(table_dict)

    assert table.name == 'my_table'
    assert table.schema == 'schema'
    assert table.columns == [
        table_structures.Column(name='col1', datatype='varchar', format_=[123], required=False),
        table_structures.Column(name='col2', datatype='integer', required=True),
        table_structures.Column(name='col3', datatype='integer', required=True,
                                default=table_structures.ColumnDefault(
                                    table_structures.ColumnDefaultType.expression,
                                    '-1')),
    ]
    assert table.foreign_keys == {
        table_structures.ForeignKey(
            column='col2', remote_table='other_table', remote_column='other_col'),
    }
    assert table.indexes == {
        table_structures.Index(columns=['col1', 'col2'])
    }
    assert table.unique_constraints == {
        table_structures.UniqueConstraint(['col3']),
    }
    assert table.primary_keys == ['col2', 'col1']


def test_table_from_sqla():
    """ Test that we take a SQL Alchemy table and make a Table structure """

    sa_table = sa.Table(
        'test_table',
        sa.MetaData(),
        sa.Column('col1', sa.Integer, primary_key=True),
        sa.Column('col2', sa.Text),
        sa.Index('my_index', 'col1', 'col2'),
        sa.ForeignKeyConstraint(['col1'], ['other_table.other_col']),
        sa.UniqueConstraint('col2'),
        sa.CheckConstraint('col1::text != col2', name='check1'),
        schema='test_schema')

    table = table_structures.Table.from_sqla(sa_table)
    assert table.name == 'test_table'
    assert table.schema == 'test_schema'
    assert table.columns == [
        table_structures.Column(name='col1', datatype='integer', format_=None, required=True),
        table_structures.Column(name='col2', datatype='text', format_=None, required=False),
    ]
    assert table.foreign_keys == {
        table_structures.ForeignKey(
            column='col1', remote_table='other_table', remote_column='other_col'),
    }
    assert table.indexes == {
        table_structures.Index(columns=['col1', 'col2'])
    }
    assert table.primary_keys == ['col1']
    assert table.unique_constraints == {
        table_structures.UniqueConstraint(['col2'])
    }
    assert table.check_constraints == {
        table_structures.CheckConstraint(
            'check1', 'col1::text != col2')
    }


def test_table_from_sqla_equality_from_to_yaml(transacted_postgresql_db):
    """
    Test that a Table structure made from a reflected SQLAlchemy table is equal to the original.
    Converts the table from and to yaml to ensure it can be serialized properly
    """
    metadata = sa.MetaData(bind=transacted_postgresql_db.connection)

    transacted_postgresql_db.connection.execute('''
        create schema schema;
        create table schema.other_table (
            other_col integer unique
        );
        create sequence schema.my_table_col6_seq increment by 1 no minvalue no maxvalue;
        create table schema.my_table (
            col1 varchar(123),
            col2 integer not null primary key references schema.other_table (other_col),
            col3 integer not null unique,
            col4 timestamp with time zone not null default now(),
            col5 varchar(1) not null default 'Y',
            -- fully defined default sequence
            col6 integer not null default nextval('schema.my_table_col6_seq'::regclass),
            -- default sequence shorthand
            col7 serial
            constraint check1 check ((col1 != 'value')),
            constraint check2 check ((col1 != 'value') and (col2 != 0))
        );
        create index idx ON schema.my_table (col1, col2);
    ''')

    table_dict = {
        'name': 'my_table',
        'schema': 'schema',
        'columns': [
            {'name': 'col1', 'datatype': 'varchar', 'format': [123], 'required': False},
            {'name': 'col2', 'datatype': 'integer', 'required': True},
            {'name': 'col3', 'datatype': 'integer', 'required': True},
            {'name': 'col4', 'datatype': 'timestamptz', 'required': True,
             'default': {'value': 'now()', 'type': 'expression'}},
            {'name': 'col5', 'datatype': 'varchar', 'format': [1], 'required': True,
             # SQLAlchemy includes explicit cast when reflecting plain string defaults
             'default': {'value': "'Y'::character varying", 'type': 'expression'}},
            {'name': 'col6', 'datatype': 'integer', 'required': True,
             'default': {'value': 'schema.my_table_col6_seq', 'type': 'sequence'}},
            {'name': 'col7', 'datatype': 'integer', 'required': True,
             'default': {'value': 'schema.my_table_col7_seq', 'type': 'sequence'}},
        ],
        'foreign_keys': [
            {'column': 'col2', 'remote_table': 'schema.other_table', 'remote_column': 'other_col'}
        ],
        'check_constraints': [
            {'name': 'check1', 'check': "(col1 != 'value')"},
            {'name': 'check2', 'check': "(col1 != 'value') and (col2 != 0)"}
        ],
        'indexes': [{'columns': ['col1', 'col2']}],
        'unique_constraints': [{'columns': ['col3']}],
        'primary_keys': ['col2']
    }
    orig_table = table_structures.Table.from_dict(table_dict)

    reflected_sa = sa.Table('my_table', metadata,
                            schema='schema',
                            autoload=True, autoload_with=transacted_postgresql_db.connection)
    reflected_table_yaml = table_structures.Table.from_sqla(reflected_sa).to_yaml()
    reflected_table = table_structures.Table.from_yaml(reflected_table_yaml)
    assert reflected_table == orig_table


def test_geography_reflection(transacted_postgresql_db):
    """
    Test that a geography column can be properly reflected
    """
    metadata = sa.MetaData(bind=transacted_postgresql_db.connection)

    transacted_postgresql_db.connection.execute('''
        create schema schema;
        create table schema.my_table (
            zip_geocode geography(Point,4326)
        );
    ''')

    table_dict = {
        'name': 'my_table',
        'schema': 'schema',
        'columns': [
            {'name': 'zip_geocode', 'datatype': 'geography', 'format': ['point', 4326]}
        ]
    }
    orig_table = table_structures.Table.from_dict(table_dict)

    reflected_sa = sa.Table('my_table', metadata,
                            schema='schema',
                            autoload=True, autoload_with=transacted_postgresql_db.connection)
    reflected_table_yaml = table_structures.Table.from_sqla(reflected_sa).to_yaml()
    reflected_table = table_structures.Table.from_yaml(reflected_table_yaml)
    assert reflected_table == orig_table


def test_create_geography_column(transacted_postgresql_db):
    """
    Test that a geograph column can be created in a table when converting a dict to sqlalchemy
    """
    metadata = sa.MetaData(bind=transacted_postgresql_db.connection)

    table_dict = {
        'name': 'my_table',
        'schema': 'schema',
        'columns': [
            {'name': 'zip_geocode', 'datatype': 'geography', 'format': ['point', 4326]}
        ]
    }
    table = table_structures.Table.from_dict(table_dict)
    sqla_table = table.to_sqla(metadata)
    transacted_postgresql_db.connection.execute('CREATE SCHEMA schema;')
    sqla_table.create()


def test_table_to_dict():
    """ Test that the whole table dict structure gets
    created successfully """

    table = table_structures.Table(
        name='foo',
        schema='test',
        doc='my doc',
        columns=[
            table_structures.Column(name='id', datatype='integer', required=True, doc='id doc'),
            table_structures.Column(name='other_id', datatype='integer'),
            table_structures.Column(name='another_id', datatype='integer'),
            table_structures.Column(name='source', datatype='uuid', required=True,
                                    default=table_structures.ColumnDefault(
                                        table_structures.ColumnDefaultType.expression,
                                        'uuid_generate_v4()')),
        ],
        primary_keys=['id', 'another_id', 'other_id'],
        foreign_keys=[
            table_structures.ForeignKey('other_id', 'other', 'id'),
        ],
        indexes=[
            table_structures.Index(['other_id']),
            table_structures.Index(['other_id', 'id']),
        ],
        unique_constraints=[
            table_structures.UniqueConstraint(['another_id'], name='unique1'),
        ],
        check_constraints=[
            table_structures.CheckConstraint('check1', 'id != other_id')
        ])
    assert table.to_dict() == {
        'name': 'foo',
        'schema': 'test',
        'doc': 'my doc',
        'columns': [
            {'name': 'id', 'datatype': 'integer', 'required': True, 'doc': 'id doc'},
            {'name': 'other_id', 'datatype': 'integer', 'required': False},
            {'name': 'another_id', 'datatype': 'integer', 'required': False},
            {'name': 'source', 'datatype': 'uuid', 'required': True,
             'default': {'value': 'uuid_generate_v4()', 'type': 'expression'}},
        ],
        'foreign_keys': [
            {'column': 'other_id', 'remote_table': 'other', 'remote_column': 'id'},
        ],
        'indexes': [
            {'columns': ['other_id']},
            {'columns': ['other_id', 'id']},
        ],
        'unique_constraints': [
            {'columns': ['another_id'], 'name': 'unique1'}
        ],
        'primary_keys': ['id', 'another_id', 'other_id'],
        'check_constraints': [{'name': 'check1', 'check': 'id != other_id'}]
    }


def test_table_to_sqla():
    """ Test that we tae a full table_structures.Table """

    table_obj = table_structures.Table(
        name='a_table',
        schema='a_schema',
        columns=[
            table_structures.Column(name='col1', datatype='varchar', format_=[123],
                                    required=False),
            table_structures.Column(name='col2', datatype='integer', required=True),
            table_structures.Column(name='col3', datatype='integer'),
            table_structures.Column(name='col4', datatype='timestamptz', required=True,
                                    default=table_structures.ColumnDefault(
                                        table_structures.ColumnDefaultType.expression,
                                        'now()')),
            table_structures.Column(name='col5', datatype='bigint', required=True,
                                    default=table_structures.ColumnDefault(
                                        table_structures.ColumnDefaultType.sequence,
                                        'a_schema.a_table_col5_seq')),
        ],
        foreign_keys=[table_structures.ForeignKey('col2', 'other_table', 'other_col')],
        indexes=[table_structures.Index(['col1'])],
        primary_keys=['col2'],
        unique_constraints=[table_structures.UniqueConstraint(['col3'], name='uq_col3')],
        check_constraints=[table_structures.CheckConstraint('check1', 'col1::text != col2')])

    meta = sa.MetaData()
    sa_table = table_obj.to_sqla(metadata=meta)

    assert sa_table.name == 'a_table'
    assert sa_table.schema == 'a_schema'
    assert {col.name for col in sa_table.c} == {'col1', 'col2', 'col3', 'col4', 'col5'}
    assert [i.constraint.columns.keys() for i in sa_table.foreign_keys] == [['col2']]
    assert [j.name for i in sa_table.indexes for j in i.expressions] == ['col1']
    assert [(c.name, c.sqltext) for c in sa_table.constraints if isinstance(c, sa.CheckConstraint)]
    assert [(c.name, [col.name for col in c.columns])
            for c in sa_table.constraints if isinstance(c, sa.UniqueConstraint)] == \
           [('uq_col3', ['col3'])]
    assert sa_table.c.col2.primary_key
    assert sa_table.c.col4.server_default
    assert str(sa_table.c.col4.server_default.arg) == 'now()'
    assert sa_table.c.col5.server_default
    assert isinstance(sa_table.c.col5.server_default, sa.DefaultClause)
    assert isinstance(sa_table.c.col5.server_default.arg, sa_func.next_value)
    assert sa_table.c.col5.server_default.arg.sequence.name == 'a_table_col5_seq'
    assert sa_table.c.col5.server_default.arg.sequence.schema == 'a_schema'
    assert sa_table.metadata == meta  # assert it's the meta I gave it

    # test without indexes
    sa_table_no_indexes = table_obj.to_sqla(metadata=sa.MetaData(), include_indexes=False)
    assert [j.name for i in sa_table_no_indexes.indexes for j in i.expressions] == []


# pylint: disable=pointless-statement,expression-not-assigned
def test_table_or():
    table = table_structures.Table(
        name='foo',
        schema='test',
        columns=[
            table_structures.Column(name='id', datatype='integer', required=True),
            table_structures.Column(name='other_id', datatype='integer'),
            table_structures.Column(name='another_id', datatype='integer'),
        ],
        primary_keys=['id'],
        foreign_keys=[
            table_structures.ForeignKey('other_id', 'other', 'id'),
        ],
        indexes=[
            table_structures.Index(['other_id']),
        ],
        unique_constraints=[
            table_structures.UniqueConstraint('other_id'),
        ],
        check_constraints=[
            table_structures.CheckConstraint('check1', 'id != other_id'),
        ])
    assert (table | None) == table
    assert (None | table) == table

    other = copy.deepcopy(table)
    other.columns = other.columns[:-1] + [
        table_structures.Column(name='creation_date', datatype='date')]
    merged = table | other
    merged_columns = {col.name for col in merged.columns}
    assert merged_columns == {'id', 'other_id', 'another_id', 'creation_date'}

    other = copy.deepcopy(table)
    other.indexes = set()
    merged = table | other
    assert merged.indexes == table.indexes

    other = copy.deepcopy(table)
    other.foreign_keys = set()
    merged = table | other
    assert not merged.foreign_keys

    other = copy.deepcopy(table)
    other.foreign_keys = {table_structures.ForeignKey('other_id', 'some_other_table', 'id')}
    merged = table | other
    assert not merged.foreign_keys

    other = copy.deepcopy(table)
    other.unique_constraints = {table_structures.UniqueConstraint('another_id')}
    merged = table | other
    assert not merged.unique_constraints

    other = copy.deepcopy(table)
    other.check_constraints = {table_structures.CheckConstraint('check2', 'other_id != 1')}
    merged = table | other
    assert not merged.check_constraints

    other = copy.deepcopy(table)
    other.columns[0].datatype = 'float'
    with pytest.raises(table_structures.MergeError):
        table | other

    other = copy.deepcopy(table)
    other.primary_keys = ['other_id']
    with pytest.raises(table_structures.MergeError):
        table | other

    other = copy.deepcopy(table)
    other.name = 'bar'
    with pytest.raises(table_structures.MergeError):
        table | other

    with pytest.raises(TypeError):
        table | object()


def test_no_column_given():
    """ Test that we raise a NoDataType when no data type
    is given in the spec. """

    table_metadata = {
        'name': 'pktable',
        'schema': 'test_schema',
        'columns': [
            {'name': 'id'},  # missing datatype
            {'name': 'somestr', 'datatype': 'text'},
        ],
        'primary_keys': 'id',
    }
    with pytest.raises(table_structures.NoDataType):
        table_structures.Table.from_dict(table_metadata)


def test_create_table_primary_key(transacted_postgresql_db):
    """ Test that we create the primary keys correctly """

    table_metadata = {
        'name': 'pktable',
        'schema': 'test_schema',
        'columns': [
            {'name': 'id', 'datatype': 'integer'},
            {'name': 'somestr', 'datatype': 'text'},
        ],
        'primary_keys': ['id'],
    }
    sa_meta = sa.MetaData()
    table = table_structures.Table.from_dict(table_metadata)
    transacted_postgresql_db.connection.execute('CREATE SCHEMA test_schema')
    sa_table = table.to_sqla(metadata=sa_meta)
    sa_table.create(transacted_postgresql_db.connection)
    assert sa_table.c.id.primary_key


def test_create_table_primary_key_multiple(transacted_postgresql_db):
    table_metadata = {
        'name': 'pktable',
        'schema': 'test_schema',
        'columns': [
            {'name': 'composite_key_1', 'datatype': 'text'},
            {'name': 'composite_key_2', 'datatype': 'text'},
        ],
        'primary_keys': ['composite_key_1', 'composite_key_2'],
    }
    sa_meta = sa.MetaData()
    table = table_structures.Table.from_dict(table_metadata)
    transacted_postgresql_db.connection.execute('CREATE SCHEMA test_schema')
    sa_table = table.to_sqla(metadata=sa_meta)
    sa_table.create(transacted_postgresql_db.connection)
    assert sa_table.c.composite_key_1.primary_key
    assert sa_table.c.composite_key_2.primary_key


def test_create_table_required_column(transacted_postgresql_db):
    table_metadata = {
        'name': 'reqcoltable',
        'schema': 'test_schema',
        'columns': [
            {'name': 'required', 'datatype': 'text', 'required': True},
            {'name': 'optional', 'datatype': 'text', 'required': False},
            {'name': 'other', 'datatype': 'text'},
        ],
    }
    sa_meta = sa.MetaData()
    table = table_structures.Table.from_dict(table_metadata)
    transacted_postgresql_db.connection.execute('CREATE SCHEMA test_schema')
    sa_table = table.to_sqla(metadata=sa_meta)
    sa_table.create(transacted_postgresql_db.connection)
    assert not sa_table.c.required.nullable
    assert sa_table.c.optional.nullable
    assert sa_table.c.other.nullable


# TABLE SET

def test_tableset_from_dict():
    """ Test that we pull up a tableset correctly """

    test_dict = {
        'tables': [
            {'name': 'foo', 'schema': 'test', 'columns': []},
            {'name': 'bar', 'schema': 'test', 'columns': []},
        ],
    }
    table_set = table_structures.TableSet.from_dict(test_dict)
    assert len(table_set.tables) == 2
    assert {t.name for t in table_set.tables} == {'foo', 'bar'}


def test_tableset_from_sqla():
    """ Test that we can take a list of sqlalchemy tables and
    return a table set containing table_structures.Table for
    each table """

    table_1 = sa.Table('my_table1', sa.MetaData(), schema='test')
    table_2 = sa.Table('my_table2', sa.MetaData(), schema='test')
    table_set = table_structures.TableSet.from_sqla([table_1, table_2])
    assert table_set.to_dict() == {
        'tables': [
            {'name': 'my_table1', 'schema': 'test', 'columns': []},
            {'name': 'my_table2', 'schema': 'test', 'columns': []},
        ],
    }


def test_tableset_to_dict():
    """ Test that we create the dict for thet tableset """

    table_set = table_structures.TableSet(tables=[
        table_structures.Table(name='foo', schema='test'),
        table_structures.Table(name='bar', schema='test'),
    ])
    assert table_set.to_dict() == {
        'tables': [
            {'name': 'foo', 'schema': 'test', 'columns': []},
            {'name': 'bar', 'schema': 'test', 'columns': []},
        ],
    }
    assert (table_set | None) == table_set
    assert (None | table_set) == table_set
    assert (table_set | table_set) == table_set


def test_tableset_to_sqla():
    """ Test that we take a tableset and return a list of sqlalchemy tables """

    table_set = table_structures.TableSet(tables=[
        table_structures.Table(name='foo', schema='test'),
        table_structures.Table(name='bar', schema='test'),
    ])
    sa_tables = table_set.to_sqla()

    assert all(isinstance(t, sa.Table) for t in sa_tables)
    assert {t.name for t in sa_tables} == {'foo', 'bar'}


def test_tableset_unique_constraints():
    """ Test that we properly add unique constraints to foreign
    key targets when calling to_sqla() """

    table_set = table_structures.TableSet.from_dict(
        {
            'tables': [{
                'name': 'one',
                'schema': 'test_schema',
                'columns': [
                    {'name': 'id', 'datatype': 'integer'},
                    {'name': 'somestr', 'datatype': 'text'},
                ],
            }, {
                'name': 'two',
                'schema': 'test_schema',
                'columns': [
                    {'name': 'one_id', 'datatype': 'integer'},
                    {'name': 'otherstr', 'datatype': 'text'},
                ],
                'foreign_keys': [{
                    'column': 'one_id',
                    'remote_table': 'test_schema.one',
                    'remote_column': 'id',
                }],
                'indexes': [{'columns': ['one_id']}],
            }],
        }
    )
    # make sure it's not there if we don't want it
    tables = table_set.to_sqla(add_unique_constraints=False)
    table_one = [i for i in tables if i.name == 'one'][0]
    assert [i for i in table_one.constraints
            if isinstance(i, sa.UniqueConstraint)] == []

    # make sure unique constraint is there by default
    tables = table_set.to_sqla(metadata=sa.MetaData())
    table_one = [i for i in tables if i.name == 'one'][0]
    table_one_unique_constraint = [
        i for i in table_one.constraints if isinstance(i, sa.UniqueConstraint)
    ][0]
    assert table_one_unique_constraint.columns.keys() == ['id']



from setuptools import setup
import sys

# Check the Python version manually because pip < 9.0 doesn't check it for us.
if sys.version_info < (3, 5):
    raise RuntimeError('Unsupported version of Python: ' + sys.version)

setup(
    setup_requires=['pbr'],
    pbr=True,
)




""" Test that we print yaml out in a pretty way
with the keys ordered by the __slots__ order """
import yaml

from yalchemy import table_structures


class CustomTestLoader(yaml.SafeLoader):  # pylint: disable=too-many-ancestors
    def __init__(self, stream):
        """
        :param stream: data stream
        """
        super().__init__(stream)
        self.add_constructor('!my_custom_tag', CustomTestLoader.custom_tag)

    def custom_tag(self, node):  # pylint: disable=no-self-use
        assert isinstance(node, yaml.ScalarNode)
        assert isinstance(node.value, str)

        return '--{}--'.format(node.value)


TEST_YAML_PRETTY = """
tables:
  - name: test_table
    schema: test_schema
    columns:
      - {name: my_col, datatype: text, required: false}
"""

REVERSE_TEST_YAML_PRETTY = """
tables:
  - columns:
      - {required: false, datatype: text, name: my_col}
    schema: test_schema
    name: test_table
"""

TEST_YAML_CUSTOM_TAG = """
tables:
  - name: test_table
    schema: test_schema
    columns:
      - {name: my_col, datatype: text, required: false}
    doc: !my_custom_tag 'My Description'
"""
REVERSE_TEST_YAML_CUSTOM_TAG = """
tables:
  - name: test_table
    schema: test_schema
    columns:
      - {name: my_col, datatype: text, required: false}
    doc: --My Description--
"""


def test_pretty_yaml_printing():
    """ Test a simple case that we correctly
    orders a table and column by it's slots  """
    col = table_structures.Column('my_col', 'text')
    table = table_structures.Table(name='test_table', schema='test_schema', columns=[col])
    table_set = table_structures.TableSet(tables=[table])
    assert table_set.to_yaml() == TEST_YAML_PRETTY.lstrip()

    # test with reversing the __slots__ order in the table and column
    table.__slots__ = table.__slots__[::-1]
    col.__slots__ = col.__slots__[::-1]
    assert table_set.to_yaml() == REVERSE_TEST_YAML_PRETTY.lstrip()


def test_custom_yaml_loader():
    """
    Test the ability to use a custom PyYAML loader
    """
    tables = table_structures.TableSet.from_yaml(
        TEST_YAML_CUSTOM_TAG.lstrip(),
        loader_factory=CustomTestLoader,
    )
    assert isinstance(tables, table_structures.TableSet)

    assert tables.to_yaml() == REVERSE_TEST_YAML_CUSTOM_TAG.lstrip()
